#include <meshoptimizer.h>
#pragma warning(push)
#pragma warning(disable : 4530)
#define TINYOBJLOADER_IMPLEMENTATION
#include <tiny_obj_loader.h>
#pragma warning(pop)

#include <unordered_map>
#include <vector>

#include "meshlet_generator.hpp"
#include <runtime/mem_arena.h>

meshlet_generator::meshlet_generator()
    : m_meshes(nullptr), m_num_meshes(0u), m_meshlets(nullptr), m_num_meshlets(0u) {
}

meshlet_generator::~meshlet_generator() {
    Release();
}

void meshlet_generator::Release() {
    for (uint32_t i = 0; i < m_num_meshes; ++i) {
        delete m_meshes[i].vertices;
        delete m_meshes[i].indices;
    }
    delete m_meshes;
    m_meshes     = nullptr;
    m_num_meshes = 0;

    for (uint32_t i = 0; i < m_num_meshlets; ++i) {
        delete m_meshlets[i].indices;
        delete m_meshlets[i].vertices;
    }
    delete m_meshlets;
    m_meshlets     = nullptr;
    m_num_meshlets = 0;
}

rt_result meshlet_generator::LoadObj(const char *path) {
    tinyobj::ObjReaderConfig config;
    config.mtl_search_path = "./";
    config.triangulate     = true;
    tinyobj::ObjReader reader;

    if (!reader.ParseFromFile(path, config)) {
        if (!reader.Error().empty()) {
            rtReportError("OBJ", "Error: %s", reader.Error().c_str());
        }
        return RT_UNKNOWN_ERROR;
    }

    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);

    auto &attrib = reader.GetAttrib();
    auto &shapes = reader.GetShapes();
    // auto &materials = reader.GetMaterials();

    mesh_data *meshes = new mesh_data[shapes.size()];
    RT_VERIFY(meshes);

    for (size_t shape_idx = 0; shape_idx < shapes.size(); ++shape_idx) {

        rt_temp_arena rewind = rtBeginTempArena(temp.arena);

        size_t num_faces           = shapes[shape_idx].mesh.num_face_vertices.size();
        vertex *unindexed_vertices = RT_ARENA_PUSH_ARRAY(temp.arena, vertex, num_faces * 3);

        size_t index_offset = 0;
        for (size_t face_idx = 0; face_idx < shapes[shape_idx].mesh.num_face_vertices.size();
             ++face_idx) {
            size_t fv = shapes[shape_idx].mesh.num_face_vertices[face_idx];
            for (size_t vert_idx = 0; vert_idx < fv; ++vert_idx) {
                tinyobj::index_t idx = shapes[shape_idx].mesh.indices[index_offset + vert_idx];
                tinyobj::real_t vx   = attrib.vertices[3 * idx.vertex_index];
                tinyobj::real_t vy   = attrib.vertices[3 * idx.vertex_index + 1];
                tinyobj::real_t vz   = attrib.vertices[3 * idx.vertex_index + 2];

                unindexed_vertices[index_offset + vert_idx].vx = vx;
                unindexed_vertices[index_offset + vert_idx].vy = vy;
                unindexed_vertices[index_offset + vert_idx].vz = vz;
            }

            index_offset += fv;
        }

        // Generate a non-redundant index buffer
        size_t num_indices  = num_faces * 3;
        unsigned int *remap = RT_ARENA_PUSH_ARRAY(temp.arena, unsigned int, num_indices);
        size_t num_vertices = meshopt_generateVertexRemap(remap,
                                                          nullptr,
                                                          num_indices,
                                                          unindexed_vertices,
                                                          num_faces * 3,
                                                          sizeof(vertex));

        meshes[shape_idx].num_indices  = (uint32_t)num_indices;
        meshes[shape_idx].indices      = new uint32_t[num_indices];
        meshes[shape_idx].num_vertices = (uint32_t)num_vertices;
        meshes[shape_idx].vertices     = new vertex[num_vertices];

        meshopt_remapIndexBuffer(meshes[shape_idx].indices, nullptr, num_indices, remap);
        meshopt_remapVertexBuffer(meshes[shape_idx].vertices,
                                  unindexed_vertices,
                                  num_faces * 3,
                                  sizeof(vertex),
                                  remap);

        rtEndTempArena(rewind);
    }

    m_num_meshes = (uint32_t)shapes.size();
    m_meshes     = meshes;

    return RT_SUCCESS;
}

rt_result meshlet_generator::RunFlat(uint32_t mesh_idx) {
    const float cone_weight = 0.5f;

    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);

    size_t max_meshlets       = meshopt_buildMeshletsBound(m_meshes[mesh_idx].num_indices,
                                                     L0_MESHLET_VERTICES,
                                                     L0_MESHLET_TRIANGLES);
    meshopt_Meshlet *meshlets = RT_ARENA_PUSH_ARRAY(temp.arena, meshopt_Meshlet, max_meshlets);
    unsigned int *meshlet_vertices =
        RT_ARENA_PUSH_ARRAY(temp.arena, unsigned int, max_meshlets *L0_MESHLET_VERTICES);
    unsigned char *meshlet_triangles =
        RT_ARENA_PUSH_ARRAY(temp.arena, unsigned char, max_meshlets *L0_MESHLET_INDICES);

    size_t meshlet_count = meshopt_buildMeshlets(meshlets,
                                                 meshlet_vertices,
                                                 meshlet_triangles,
                                                 m_meshes[mesh_idx].indices,
                                                 m_meshes[mesh_idx].num_indices,
                                                 &m_meshes[mesh_idx].vertices[0].vx,
                                                 m_meshes[mesh_idx].num_vertices,
                                                 sizeof(vertex),
                                                 L0_MESHLET_VERTICES,
                                                 L0_MESHLET_TRIANGLES,
                                                 cone_weight);

    m_meshlets     = new meshlet[meshlet_count];
    m_num_meshlets = (uint32_t)meshlet_count;

    for (size_t i = 0; i < meshlet_count; ++i) {
        meshopt_optimizeMeshlet(&meshlet_vertices[meshlets[i].vertex_offset],
                                &meshlet_triangles[meshlets[i].triangle_offset],
                                meshlets[i].triangle_count,
                                meshlets[i].vertex_count);

        meshopt_Bounds bounds =
            meshopt_computeMeshletBounds(&meshlet_vertices[meshlets[i].vertex_offset],
                                         &meshlet_triangles[meshlets[i].triangle_offset],
                                         meshlets[i].triangle_count,
                                         &m_meshes[mesh_idx].vertices[0].vx,
                                         m_meshes[mesh_idx].num_vertices,
                                         sizeof(vertex));

        m_meshlets[i].vertices     = new vertex[L0_MESHLET_VERTICES];
        m_meshlets[i].indices      = new uint16_t[L0_MESHLET_INDICES];
        m_meshlets[i].num_vertices = meshlets[i].vertex_count;
        m_meshlets[i].num_indices  = meshlets[i].triangle_count * 3;
        memcpy(m_meshlets[i].center, bounds.center, sizeof(bounds.center));
        m_meshlets[i].radius = bounds.radius;

        for (unsigned int vert_idx = 0; vert_idx < meshlets[i].vertex_count; ++vert_idx) {
            unsigned int vert = meshlet_vertices[meshlets[i].vertex_offset + vert_idx];
            m_meshlets[i].vertices[vert_idx] = m_meshes[mesh_idx].vertices[vert];
        }

        for (unsigned int tri_idx = 0; tri_idx < meshlets[i].triangle_count; ++tri_idx) {
            m_meshlets[i].indices[tri_idx * 3 + 0] =
                (uint16_t)meshlet_triangles[meshlets[i].triangle_offset + 3 * tri_idx + 0];
            m_meshlets[i].indices[tri_idx * 3 + 1] =
                (uint16_t)meshlet_triangles[meshlets[i].triangle_offset + 3 * tri_idx + 1];
            m_meshlets[i].indices[tri_idx * 3 + 2] =
                (uint16_t)meshlet_triangles[meshlets[i].triangle_offset + 3 * tri_idx + 2];
        }
        m_meshlets[i].num_children = 0u;
    }

    return RT_SUCCESS;
}

rt_result meshlet_generator::RunHierarchical(uint32_t mesh_idx, uint32_t *out_root) {
    const float cone_weight = 0.25f;

    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);

    // Convert meshopt meshlets into our meshlets, each with its own vertex and index buffer
    std::vector<meshlet> meshlets;
    {
        size_t max_meshlets = meshopt_buildMeshletsBound(m_meshes[mesh_idx].num_indices,
                                                         L0_MESHLET_VERTICES,
                                                         L0_MESHLET_TRIANGLES);
        meshopt_Meshlet *meshopt_meshlets =
            RT_ARENA_PUSH_ARRAY(temp.arena, meshopt_Meshlet, max_meshlets);
        unsigned int *meshlet_vertices =
            RT_ARENA_PUSH_ARRAY(temp.arena, unsigned int, max_meshlets *L0_MESHLET_VERTICES);
        unsigned char *meshlet_triangles =
            RT_ARENA_PUSH_ARRAY(temp.arena, unsigned char, max_meshlets *L0_MESHLET_INDICES);

        uint32_t meshlet_count = (uint32_t)meshopt_buildMeshlets(meshopt_meshlets,
                                                                 meshlet_vertices,
                                                                 meshlet_triangles,
                                                                 m_meshes[mesh_idx].indices,
                                                                 m_meshes[mesh_idx].num_indices,
                                                                 &m_meshes[mesh_idx].vertices[0].vx,
                                                                 m_meshes[mesh_idx].num_vertices,
                                                                 sizeof(vertex),
                                                                 L0_MESHLET_VERTICES,
                                                                 L0_MESHLET_TRIANGLES,
                                                                 cone_weight);

        for (size_t i = 0; i < meshlet_count; ++i) {
            meshopt_optimizeMeshlet(&meshlet_vertices[meshopt_meshlets[i].vertex_offset],
                                    &meshlet_triangles[meshopt_meshlets[i].triangle_offset],
                                    meshopt_meshlets[i].triangle_count,
                                    meshopt_meshlets[i].vertex_count);

            meshopt_Bounds bounds = meshopt_computeMeshletBounds(
                &meshlet_vertices[meshopt_meshlets[i].vertex_offset],
                &meshlet_triangles[meshopt_meshlets[i].triangle_offset],
                meshopt_meshlets[i].triangle_count,
                &m_meshes[mesh_idx].vertices[0].vx,
                m_meshes[mesh_idx].num_vertices,
                sizeof(vertex));

            meshlet meshlet;
            meshlet.vertices     = new vertex[L0_MESHLET_VERTICES];
            meshlet.indices      = new uint16_t[L0_MESHLET_INDICES];
            meshlet.num_vertices = meshopt_meshlets[i].vertex_count;
            meshlet.num_indices  = meshopt_meshlets[i].triangle_count * 3;
            memcpy(meshlet.center, bounds.center, sizeof(bounds.center));
            meshlet.radius = bounds.radius;

            for (unsigned int vert_idx = 0; vert_idx < meshopt_meshlets[i].vertex_count;
                 ++vert_idx) {
                unsigned int vert = meshlet_vertices[meshopt_meshlets[i].vertex_offset + vert_idx];
                meshlet.vertices[vert_idx] = m_meshes[mesh_idx].vertices[vert];
            }

            for (unsigned int tri_idx = 0; tri_idx < meshopt_meshlets[i].triangle_count;
                 ++tri_idx) {
                meshlet.indices[tri_idx * 3 + 0] = (uint16_t)
                    meshlet_triangles[meshopt_meshlets[i].triangle_offset + 3 * tri_idx + 0];
                meshlet.indices[tri_idx * 3 + 1] = (uint16_t)
                    meshlet_triangles[meshopt_meshlets[i].triangle_offset + 3 * tri_idx + 1];
                meshlet.indices[tri_idx * 3 + 2] = (uint16_t)
                    meshlet_triangles[meshopt_meshlets[i].triangle_offset + 3 * tri_idx + 2];
            }

            meshlet.num_children = 0u;

            meshlets.push_back(meshlet);
        }
    }

    // We now have a flat list of meshlets -> the highest lod ones
    // We now combine (up to 8) meshlets into one to generate the next hierarchy level
    // Repeat until we only have 1 meshlet left
    std::vector<uint32_t> unprocessed;
    unprocessed.reserve(meshlets.size());
    for (uint32_t i = 0; i < meshlets.size(); ++i) {
        unprocessed.push_back(i);
    }
    std::vector<uint32_t> next_level;
    next_level.reserve((unprocessed.size() + 7) / 8);
    while (!unprocessed.empty()) {
        uint32_t first = unprocessed.back();
        unprocessed.pop_back();

        // Find the 7 closest center points
        uint32_t closest[7];
        float distances[7];
        unsigned int closest_count = 0u;

        float first_center[3];
        memcpy(first_center, meshlets[first].center, sizeof(float) * 3);

        for (uint32_t i = 0; i < unprocessed.size(); ++i) {
            float center[3];
            memcpy(center, meshlets[unprocessed[i]].center, sizeof(float) * 3);

            float dist = sqrtf((center[0] - first_center[0]) * (center[0] - first_center[0]) +
                               (center[1] - first_center[1]) * (center[1] - first_center[1]) +
                               (center[2] - first_center[2]) * (center[2] - first_center[2]));

            if (closest_count == RT_ARRAY_COUNT(closest)) {
                // Check if we are closer than one of the other candidates
                for (unsigned int j = 0; j < closest_count; ++j) {
                    uint32_t highest_idx = UINT_MAX;
                    float highest_dist   = dist;
                    if (dist < distances[j]) {
                        if (distances[j] > highest_dist) {
                            highest_dist = distances[j];
                            highest_idx  = j;
                        }
                    }
                    if (highest_idx < RT_ARRAY_COUNT(closest)) {
                        const uint32_t replaced = closest[highest_idx];
                        distances[j]            = dist;
                        closest[j]              = unprocessed[i];
                        unprocessed.push_back(replaced);
                    }
                }
            } else {
                closest[closest_count]   = unprocessed[i];
                distances[closest_count] = dist;
                closest_count++;
                unprocessed.erase(unprocessed.begin() + i);
            }
        }

        // Combine into a new meshlet
        // vertex *vertices = new vertex[MESHLET_VERTICES * 8];

        vertex *in_vertices  = new vertex[L0_MESHLET_VERTICES * 8];
        uint32_t *in_indices = new uint32_t[L0_MESHLET_INDICES * 8];
        memcpy(in_vertices,
               meshlets[first].vertices,
               sizeof(vertex) * meshlets[first].num_vertices);
        size_t at_vert = meshlets[first].num_vertices;
        for (unsigned int i = 0; i < meshlets[first].num_indices; ++i)
            in_indices[i] = static_cast<uint32_t>(meshlets[first].indices[i]);
        size_t at_idx = meshlets[first].num_indices;
        for (unsigned int i = 0; i < closest_count; ++i) {
            memcpy(&in_vertices[at_vert],
                   meshlets[closest[i]].vertices,
                   sizeof(vertex) * meshlets[closest[i]].num_vertices);
            at_vert += meshlets[closest[i]].num_vertices;

            for (unsigned int j = 0; j < meshlets[closest[i]].num_indices; ++j)
                in_indices[at_idx + j] = static_cast<uint32_t>(meshlets[closest[i]].indices[j]);
            at_idx += meshlets[closest[i]].num_indices;
        }

        const float target_error =
            0.10f; // Acept 10% error. We are simplifying meshlets, the extents are not very large
        const size_t target_index_count     = L0_MESHLET_INDICES;
        const unsigned int simplify_options = meshopt_SimplifyLockBorder;
        const float threshold               = 1.f / 8.f;
        std::vector<unsigned int> lod(UINT16_MAX);
        float lod_error        = 0.f;
        size_t out_index_count = meshopt_simplify(&lod[0],
                                                  in_indices,
                                                  at_idx,
                                                  &in_vertices[0].vx,
                                                  at_vert,
                                                  sizeof(vertex),
                                                  target_index_count,
                                                  target_error,
                                                  simplify_options,
                                                  &lod_error);
        lod.resize(out_index_count);
        if (out_index_count < UINT16_MAX) {
            rtLog("EXP", "Yay");
        } else {
            rtReportError("EXP", "Nay");
        }

        std::unordered_map<unsigned int, uint16_t> index_remap;
        uint16_t *index_buffer = new uint16_t[out_index_count];
        std::vector<vertex> merged_vertices;
        merged_vertices.reserve(out_index_count);
        for (size_t i = 0; i < out_index_count; ++i) {
            if (index_remap.find(lod[i]) != index_remap.end()) {
                index_buffer[i] = index_remap[lod[i]];
            } else {
                uint16_t index = (uint16_t)merged_vertices.size();
                merged_vertices.push_back(in_vertices[lod[i]]);
                index_buffer[i]     = index;
                index_remap[lod[i]] = index;
            }
        }

        meshlet merged      = {};
        merged.num_children = closest_count + 1;
        merged.children[0]  = first;
        memcpy(merged.children, closest, sizeof(uint32_t) * closest_count);
        merged.indices     = index_buffer;
        merged.num_indices = (uint32_t)out_index_count;
        merged.vertices    = new vertex[merged_vertices.size()];
        merged.num_vertices = (uint32_t)merged_vertices.size();
        memcpy(merged.vertices, merged_vertices.data(), sizeof(vertex) * merged_vertices.size());

        // Determine center and bounds


        meshlets.push_back(merged);
        next_level.push_back((unsigned int)meshlets.size() - 1);
    }
    return RT_SUCCESS;
}